import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.lines import Line2D
import numpy as np
from os.path import join


def init_figure():
    h = plt.figure(figsize=(11.7, 8.3))
    gs = GridSpec(17, 3)
    gs.update(wspace=0.25)
    axr_sum = plt.subplot(gs[0, 0])
    axr_sum.grid(color='gray', linestyle='--', linewidth=0.4, axis='x')
    axr = plt.subplot(gs[1:, 0])
    axr.grid(color='gray', linestyle='--', linewidth=0.4, axis='x')
    axt_sum = plt.subplot(gs[0, 1])
    axt_sum.grid(color='gray', linestyle='--', linewidth=0.4, axis='x')
    axt = plt.subplot(gs[1:, 1])
    axt.grid(color='gray', linestyle='--', linewidth=0.4, axis='x')
    axb = plt.subplot(gs[1:, 2])
    axb.grid(color='gray', linestyle='--', linewidth=0.4, axis='x')
    return h, axr, axt, axb, axr_sum, axt_sum


def plot_waves(axr, axt, axb, axr_sum, axt_sum, stadata, enf=3):
    bound = np.zeros(stadata.rflength)
    for i in range(stadata.ev_num):
        datar = stadata.datar[i] * enf + (i + 1)
        datat = stadata.datat[i] * enf + (i + 1)
        # axr.plot(time_axis, stadata.datar[i], linewidth=0.2, color='black')
        axr.fill_between(stadata.time_axis, datar, bound + i+1, where=datar > i+1, facecolor='red',
                         alpha=0.7)
        axr.fill_between(stadata.time_axis, datar, bound + i+1, where=datar < i+1, facecolor='blue',
                         alpha=0.7)
        # axt.plot(time_axis, stadata.datat[i], linewidth=0.2, color='black')
        axt.fill_between(stadata.time_axis, datat, bound + i + 1, where=datat > i+1, facecolor='red',
                         alpha=0.7)
        axt.fill_between(stadata.time_axis, datat, bound + i + 1, where=datat < i+1, facecolor='blue',
                         alpha=0.7)
    datar = np.mean(stadata.datar, axis=0)
    datar /= np.max(datar)
    datat = np.mean(stadata.datat, axis=0)
    datat /= np.max(datar)
    axr_sum.fill_between(stadata.time_axis, datar, bound, where=datar > 0, facecolor='red', alpha=0.7)
    axr_sum.fill_between(stadata.time_axis, datar, bound, where=datar < 0, facecolor='blue', alpha=0.7)
    axt_sum.fill_between(stadata.time_axis, datat, bound, where=datat > 0, facecolor='red', alpha=0.7)
    axt_sum.fill_between(stadata.time_axis, datat, bound, where=datat < 0, facecolor='blue', alpha=0.7)
    axb.scatter(stadata.bazi, np.arange(stadata.ev_num) + 1, s=7)


def set_fig(axr, axt, axb, axr_sum, axt_sum, stadata, station, xmin=-2, xmax=30, comp='R'):
    y_range = np.arange(stadata.ev_num) + 1
    x_range = np.arange(0, xmax+2, 2)
    space = 2

    # set axr
    axr.set_xlim(xmin, xmax)
    axr.set_xticks(x_range)
    axr.set_xticklabels(x_range, fontsize=8)
    axr.set_ylim(0, stadata.ev_num + space)
    axr.set_yticks(y_range)
    axr.set_yticklabels(stadata.event, fontsize=5)
    axr.set_xlabel('Time after P (s)', fontsize=13)
    axr.set_ylabel('Event', fontsize=13)
    axr.add_line(Line2D([0, 0], axr.get_ylim(), color='black'))

    # set axr_sum
    axr_sum.set_title('{} components ({})'.format(comp, station), fontsize=16)
    axr_sum.set_xlim(xmin, xmax)
    axr_sum.set_xticks(x_range)
    axr_sum.set_xticklabels([])
    axr_sum.set_ylim(-0.5, 1.25)
    axr_sum.set_yticks([0.375])
    axr_sum.set_yticklabels(['Sum'], fontsize=8)
    axr_sum.tick_params(axis='y', left=False)
    axr_sum.add_line(Line2D([0, 0], axr_sum.get_ylim(), color='black'))

    # set axt
    axt.set_xlim(xmin, xmax)
    axt.set_xticks(x_range)
    axt.set_xticklabels(x_range, fontsize=8)
    axt.set_ylim(0, stadata.ev_num + space)
    axt.set_yticks(y_range)
    bazi = ['{:.1f}'.format(ba) for ba in stadata.bazi]
    axt.set_yticklabels(bazi, fontsize=5)
    axt.set_xlabel('Time after P (s)', fontsize=13)
    axt.set_ylabel(r'Back-azimuth ($\circ$)', fontsize=13)
    axt.add_line(Line2D([0, 0], axt.get_ylim(), color='black'))

    # set axt_sum
    axt_sum.set_title('T components ({})'.format(station), fontsize=16)
    axt_sum.set_xlim(xmin, xmax)
    axt_sum.set_xticks(x_range)
    axt_sum.set_xticklabels([])
    axt_sum.set_ylim(-0.5, 1.25)
    axt_sum.set_yticks([0.375])
    axt_sum.set_yticklabels(['Sum'], fontsize=8)
    axt_sum.tick_params(axis='y', left=False)
    axt_sum.add_line(Line2D([0, 0], axt_sum.get_ylim(), color='black'))

    # set axb
    axb.set_xlim(0, 360)
    axb.set_xticks(np.linspace(0, 360, 7))
    axb.set_xticklabels(np.linspace(0, 360, 7, dtype='i'), fontsize=8)
    axb.set_ylim(0, stadata.ev_num + space)
    axb.set_yticks(y_range)
    axb.set_yticklabels(y_range, fontsize=5)
    axb.set_xlabel(r'Back-azimuth ($\circ$)', fontsize=13)


def plotrt(rfsta, enf=3, out_path='./', key='bazi', outformat='g', xlim=[-2, 30], show=False):
    """Plot PRFs with R and T components

    :param rfpath: Path to PRFs
    :type rfpath: str
    :param enf: The enlarge factor, defaults to 3
    :type enf: int, optional
    :param out_path: The output path, defaults to current directory
    :type out_path: str, optional
    :param outformat: File format of the image file, g as \'png\', f as \'pdf\', defaults to 'g'
    :type outformat: str, optional
    """
    h, axr, axt, axb, axr_sum, axt_sum = init_figure()
    rfsta.sort(key)
    plot_waves(axr, axt, axb, axr_sum, axt_sum, rfsta, enf=enf)
    set_fig(axr, axt, axb, axr_sum, axt_sum, rfsta, rfsta.staname, xmin=xlim[0], xmax=xlim[1], comp=rfsta.comp)
    if outformat is None and not show:
        return h
    elif outformat == 'g':
        h.savefig(join(out_path, rfsta.staname+'_RT_{}order_{:.1f}.png'.format(key, rfsta.f0[0])),
                       dpi=400, bbox_inches='tight')
    elif outformat == 'f':
        h.savefig(join(out_path, rfsta.staname+'_RT_{}order_{:.1f}.pdf'.format(key, rfsta.f0[0])),
                       format='pdf', bbox_inches='tight')
    if show:
        plt.show()
        return h


if __name__ == '__main__':
    pass



import numpy as np
from scipy.fftpack import ifft
from obspy.signal.util import next_pow_2
from obspy import Trace, Stream
from seispy.decon import RFTrace

ei = 0+1j

def e_inverse(omega, rho, alpha, beta, p):
    """ E_inverse (Aki & Richards, pp. 161, Eq. (5.71))

    Parameters
    ----------
    omega : _type_
        _description_
    rho : _type_
        _description_
    alpha : _type_
        _description_
    beta : _type_
        _description_
    p : _type_
        _description_
    """
    e_inv = np.zeros([4,4], dtype=complex)
    eta = np.sqrt(1.0/(beta*beta) - p*p)
    xi  = np.sqrt(1.0/(alpha*alpha) - p*p)
    bp = 1.0 - 2.0*beta*beta*p*p

    e_inv[0,0] = beta*beta*p/alpha
    e_inv[0,1] = bp/(2.0*alpha*xi)
    e_inv[0,2] = -p/(2.0*omega*rho*alpha*xi) * ei
    e_inv[0,3] = -1.0/(2.0*omega*rho*alpha) * ei
    e_inv[1,0] = bp / (2.0*beta*eta)
    e_inv[1,1] = -beta*p
    e_inv[1,2] = -1.0/(2.0*omega*rho*beta) * ei
    e_inv[1,3] = p/(2.0*omega*rho*beta*eta) * ei
    e_inv[2,0] = e_inv[0,0]
    e_inv[2,1] = - e_inv[0,1]
    e_inv[2,2] = - e_inv[0,2]
    e_inv[2,3] = e_inv[0,3]
    e_inv[3,0] = e_inv[1,0]
    e_inv[3,1] = - e_inv[1,1]
    e_inv[3,2] = - e_inv[1,2]
    e_inv[3,3] = e_inv[1,3]
    return e_inv


def propagator_sol(omega, rho, alpha, beta, p, z):
    """ propagator (Aki & Richards, pp. 398, Eq. (3) in Box 9.1)

    Parameters
    ----------
    omega : _type_
        _description_
    rho : _type_
        _description_
    alpha : _type_
        _description_
    beta : _type_
        _description_
    p : _type_
        _description_
    z : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    
    p_mat = np.zeros([4,4], dtype=complex)
    beta2 = beta*beta
    p2 = p*p
    bp = 1.0 -2.0*beta2*p2
    eta = np.sqrt(1.0/(beta2) - p2)
    xi  = np.sqrt(1.0/(alpha*alpha) - p2)
    cos_xi = np.cos(omega*xi*z)
    cos_eta = np.cos(omega*eta*z)
    sin_xi = np.sin(omega*xi*z)
    sin_eta = np.sin(omega*eta*z)

    p_mat[0,0] = 2.0*beta2*p2*cos_xi + bp*cos_eta
    p_mat[0,1] = p*( bp/xi*sin_xi - 2.0*beta2*eta*sin_eta ) * ei
    p_mat[0,2] = (p2/xi*sin_xi + eta*sin_eta)/(omega*rho)
    p_mat[0,3] = p*(-cos_xi + cos_eta)/(omega*rho) * ei  
    p_mat[1,0] = p*( 2.0*beta2*xi*sin_xi - bp/eta*sin_eta ) * ei
    p_mat[1,1] = bp*cos_xi + 2.0*beta2*p2*cos_eta
    p_mat[1,2] = p_mat[0,3]
    p_mat[1,3] = (xi*sin_xi + p2/eta*sin_eta)/(omega*rho)
    p_mat[2,0] = omega*rho*( -4.0*beta2*beta2*p2*xi*sin_xi - bp*bp/eta*sin_eta )
    p_mat[2,1] = 2.0*omega*beta2*rho*p*bp*( cos_xi - cos_eta ) * ei
    p_mat[2,2] = p_mat[0,0]
    p_mat[2,3] = p_mat[1,0]
    p_mat[3,0] = p_mat[2,1]
    p_mat[3,1] = -omega*rho*( bp*bp/xi*sin_xi + 4.0*beta2*beta2*p2*eta*sin_eta  )
    p_mat[3,2] = p_mat[0,1]  
    p_mat[3,3] = p_mat[1,1]

    return p_mat

def haskell(omega, p, nl, ipha, alpha, beta, rho, h):
    i0 = 0
    e_inv = e_inverse(omega, rho[-1], alpha[-1], beta[-1], p)
    p_mat = propagator_sol(omega, rho[i0], alpha[i0], beta[i0], p, h[i0] )
    for i in range(i0+1, nl):
        p_mat2 = propagator_sol(omega, rho[i], alpha[i], beta[i], p, h[i])
        p_mat = np.matmul(p_mat2, p_mat)
    if nl > i0+1:
        sl = np.matmul(e_inv, p_mat)
    else:
        sl = e_inv
    denom = sl[2,0] * sl[3,1] - sl[2,1] * sl[3,0]
    if ipha >= 0:
        ur = sl[3,1] / denom
        uz = - sl[3,0] / denom
    else:
        ur = - sl[2,1] / denom
        uz = sl[2,0] / denom
    return ur, uz


def fwd_seis(rayp, dt, npts, ipha, alpha, beta, rho, h):
    nlay = h.size
    npts_max = next_pow_2(npts)
    ur_freq = np.zeros(npts_max, dtype=complex)
    uz_freq = np.zeros(npts_max, dtype=complex)
    nhalf = int(npts_max / 2 + 1)
    for i in range(1, nhalf):
        omg = 2*np.pi * i / (npts_max * dt)
        ur_freq[i], uz_freq[i] = haskell(omg, rayp, nlay, ipha, 
                                         alpha, beta, rho, h)
    ur = ifft(ur_freq).real[::-1]/npts_max
    uz = -ifft(uz_freq).real[::-1]/npts_max

    return ur[0:npts], uz[0:npts]


class SynSeis():
    def __init__(self, depmod, rayp, dt, npts=2500, ipha=1, filter=None) -> None:
        """_summary_

        Parameters
        ----------
        depmod : _type_
            DepModel class
        rayp : _type_
            Ray-parameter in s/km
        dt : _type_
            Time interval
        npts : _type_
            samples of synthetic waveform
        ipha : _type_
            Specify incident wave 1 for P and -1 for S
        """
        self.depmod = depmod
        self.dt = dt
        self.npts = npts
        if not isinstance(rayp, (float, list, np.ndarray)):
            raise TypeError('The rayp should be in float, list and np.ndarray')
        if isinstance(rayp, float):
            self.rayp = [rayp]
        else:
            self.rayp = rayp
        self.ipha = ipha

    def run_fwd(self):
        """Forward modelling synthetic seismograms.

        ``SynSeis.rstream`` and ``SynSeis.zstream`` are generated as 
        radial and vertical Seismograms in ``Obspy.Stream`` type.
        """
        self.rstream = Stream()
        self.zstream = Stream()
        for _, rayp in enumerate(self.rayp):
            ur, uz = fwd_seis(rayp, self.dt, self.npts, self.ipha,
                            self.depmod.vp, self.depmod.vs, self.depmod.rho,
                            self.depmod.thickness)
            tr = Trace(data=ur)
            tr.stats.delta = self.dt
            self.rstream.append(tr)
            tr = Trace(data=uz)
            tr.stats.delta = self.dt
            self.zstream.append(tr)

    def filter(self, freqmin, freqmax, order=2, zerophase=True):
        """Apply a bandpass filter on synthetic waveforms

        Parameters
        ----------
        freqmin : float
            Minimum cut-off frequency
        freqmax : float
            maximum cut-off frequency
        order : int, optional
            Order of filter, by default 2
        zerophase : bool, optional
            whether use a zero-phase filter, by default True
        """
        for st in [self.rstream, self.zstream]:
            st.filter('bandpass', freqmin=freqmin, freqmax=freqmax,
                      corners=order, zerophase=zerophase)

    def run_deconvolution(self, pre_filt=[0.05, 2], shift=10, f0=2.0, **kwargs):
        if pre_filt is not None:
            self.filter(*pre_filt)
        rfstream = Stream()
        for i, _ in enumerate(self.rayp):
            rftr = RFTrace.deconvolute(self.rstream[i], self.zstream[i], tshift=shift,
                                       f0=f0, **kwargs)
            rfstream.append(rftr)
        return rfstream




from multiprocessing.sharedctypes import Value
from obspy.io.sac.sactrace import SACTrace
import numpy as np
from scipy.interpolate import interp1d, interpn
from scipy.signal import resample
from os.path import dirname, join, exists, isfile, abspath
from seispy.geo import skm2srad, sdeg2skm, rad2deg, latlon_from, \
                       asind, tand, srad2skm, km2deg
from seispy.psrayp import get_psrayp
from seispy.rfani import RFAni
from seispy.slantstack import SlantStack
from seispy.harmonics import Harmonics
from seispy.plotRT import plotrt as _plotrt
from seispy.plotR import plotr as _plotr
from seispy.utils import DepModel, Mod3DPerturbation
import warnings
import glob


class RFStation(object):
    def __init__(self, data_path, only_r=False, prime_comp='R'):
        """
        Class for derivative process of RFs.

        :param data_path: Path to RF data with SAC format. A finallist.dat must be in this path.
        :type data_path: str
        :param only_r: Whether reading transverse RFs, defaults to False
        :type only_r: bool, optional
        :param prime_comp: Prime component in RF filename. ``R`` or ``Q`` for PRF and ``L`` or ``Z`` for SRF
        :type prime_comp: str

        .. warning::

            This Class will be renamed to ``RFStation`` in future versions.
        """
        
        self.only_r = only_r
        self.comp = prime_comp
        self.dtype = {'names': ('event', 'phase', 'evla', 'evlo', 'evdp', 'dis', 'bazi', 'rayp', 'mag', 'f0'),
                 'formats': ('U20', 'U20', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4')}
        if not exists(data_path):
            return
        self._chech_comp()
        if isfile(data_path):
            data_path = dirname(abspath(data_path))
        evt_lsts = glob.glob(join(data_path, '*finallist.dat'))
        if len(evt_lsts) == 0:
            raise FileNotFoundError("No such *finallist.dat in the {}".format(data_path))
        elif len(evt_lsts) > 1:
            raise ValueError("More than one finallist.dat in the {}".format(data_path))
        else:
            evt_lst = evt_lsts[0]
        self.event, self.phase, self.evla, self.evlo, self.evdp, self.dis, self.bazi, self.rayp, self.mag, self.f0 = \
            np.loadtxt(evt_lst, dtype=self.dtype, unpack=True, ndmin=1)
        self.rayp = skm2srad(self.rayp)
        self.ev_num = self.evla.shape[0]
        self.read_sample(data_path)
        self.data_prime = np.empty([self.ev_num, self.rflength])
        # self.__dict__['data{}'.format(self.comp.lower())] = np.empty([self.ev_num, self.rflength])
        # eval('self.data_prime = self.data{}'.format(self.comp.lower()))
        if not only_r:
            self.datat = np.empty([self.ev_num, self.rflength])
            for _i, evt, ph in zip(range(self.ev_num), self.event, self.phase):
                sac = SACTrace.read(join(data_path, evt + '_' + ph + '_{}.sac'.format(self.comp)))
                sact = SACTrace.read(join(data_path, evt + '_' + ph + '_T.sac'))
                self.data_prime[_i] = sac.data
                self.datat[_i] = sact.data
        else:
            for _i, evt, ph in zip(range(self.ev_num), self.event, self.phase):
                sac = SACTrace.read(join(data_path, evt + '_' + ph + '_{}.sac'.format(self.comp)))
                self.data_prime[_i] = sac.data
        exec('self.data{} = self.data_prime'.format(self.comp.lower()))

    def read_sample(self, data_path):
        fname = glob.glob(join(data_path, self.event[0] + '_' + self.phase[0] + '_{}.sac'.format(self.comp)))
        if len(fname) == 0:
            raise FileNotFoundError('No such files with comp of {} in {}'.format(self.comp, data_path))
        else:
            sample_sac = SACTrace.read(fname[0])
        self.staname = '{}.{}'.format(sample_sac.knetwk, sample_sac.kstnm)
        self.stla = sample_sac.stla
        self.stlo = sample_sac.stlo
        if sample_sac.stel is None:
            self.stel = 0.
        else:
            self.stel = sample_sac.stel
        self.rflength = sample_sac.npts
        self.shift = -sample_sac.b
        self.sampling = sample_sac.delta
        self.time_axis = np.arange(self.rflength) * self.sampling - self.shift

    def init_property(self, ev_num):
        self.ev_num = ev_num
        self.event = np.array(['']*ev_num)
        self.phase = np.array(['']*ev_num)
        self.evla = np.zeros(ev_num)
        self.evlo = np.zeros(ev_num)
        self.evdp = np.zeros(ev_num)
        self.mag = np.zeros(ev_num)
        self.dis = np.zeros(ev_num)
        self.bazi = np.zeros(ev_num)
        self.rayp = np.zeros(ev_num)
        self.f0 = np.zeros(ev_num)
        self.stla = 0
        self.stlo = 0
        self.stel = 0
        self.staname = ''
        self.sampling = 0
        self.rflength = 0
        # self.data_prime = np.array([])
        # exec('self.data_prime = self.data{} = np.array([])'.format(self.comp.lower()))
        # self.__dict__['data{}'.format(self.comp.lower())] = np.array([])
        self.datat = np.array([])
        # self.data_prime = eval('self.data{}'.format(self.comp.lower()))

    @classmethod
    def read_stream(cls, stream, rayp, baz, prime_comp='R'):
        """Create RFStation instance from ``obspy.Stream``

        Parameters
        ----------
        stream : ``obspy.Stream``
            _description_
        rayp : ``numpy.ndarray`` or ``float``
            Ray-parameters in s/km.
        baz : ``numpy.ndarray`` or ``float``
            Back-azimuth
        prime_comp : str, optional
             Prime component of RF, by default 'R'
        """
        if len(stream) == 0:
            raise ValueError('No such RFTrace read')
        rfsta = cls('', only_r=True, prime_comp=prime_comp)
        ev_num = len(stream)
        if isinstance(rayp, float):
            rayp = np.ones(ev_num)*rayp
        if isinstance(baz, (float, int)):
            baz = np.ones(ev_num)*baz
        if ev_num != rayp.size or ev_num != baz.size:
            raise ValueError('Array length of rayp and baz must be the same as stream')
        rfsta.init_property(ev_num)
        try:
            rfsta.staname = '{}.{}'.format(stream[0].stats.sac.knetwk, stream[0].stats.sac.kstnm)
            rfsta.stla = stream[0].stats.sac.stla
            rfsta.stlo = stream[0].stats.sac.stlo
            rfsta.stel = stream[0].stats.sac.stel
        except:
            pass
        try:
            rfsta.shift = stream[0].stats.tshift
        except:
            rfsta.shift = -stream[0].stats.sac.b
        rfsta.sampling = stream[0].stats.delta
        rfsta.rflength = stream[0].stats.npts
        rfsta.data_prime = np.zeros([rfsta.ev_num, rfsta.rflength])
        rfsta.bazi = baz
        rfsta.rayp = skm2srad(rayp)
        rfsta.time_axis = np.arange(rfsta.rflength) * rfsta.sampling - rfsta.shift
        for i, tr in enumerate(stream):
            rfsta.event[i] = '{}'.format(i)
            try:
                rfsta.f0[i] = tr.stats.f0
            except:
                rfsta.f0[i] = tr.stats.sac.user1
            rfsta.data_prime[i] = tr.data
        exec('rfsta.data{} = rfsta.data_prime'.format(rfsta.comp.lower()))
        return rfsta

    @property
    def stel(self):
        return self._stel
    
    @stel.setter
    def stel(self, value):
        if value is None:
            self._stel = 0.
        else:
            self._stel = value/1000

    def _chech_comp(self):
        if self.comp in ['R', 'Q']:
            self.prime_phase = 'P'
        elif self.comp in ['L', 'Z']:
            self.prime_phase = 'S'
        else:
            raise ValueError('prime component should be in \'R\', \'Q\', \'L\' and \'Z\'')

    def normalize(self, method='single'):
        """Normalize amplitude of each RFs.
        :param method: Method of normalization with ``single`` and ``average`` avaliable.
                     - ``single`` for normalization with max amplitude of current RF.
                     - ``average`` for normalization with average amplitude of current station.
        :type method: str, optional
        """
        if not isinstance(method, str):
            raise TypeError('\'type\' must be string, but {} type got'.format(type(method)))
        if method == 'single':
            maxamp = np.nanmax(np.abs(self.data_prime), axis=1)
        elif method == 'average':
            amp = np.nanmax(np.abs(np.mean(self.data_prime, axis=0)))
            maxamp = np.ones(self.ev_num) * amp
        else:
            raise ValueError('\'method\' must be in \'single\' and \'average\'')
        for i in range(self.ev_num):
            self.data_prime[i] /= maxamp[i]
            exec('self.data{} = self.data_prime'.format(self.comp.lower()))
            if not self.only_r:
                self.datat[i] /= maxamp[i]

    def resample(self, dt):
        """Resample RFs with specified dt


        Parameters
        ----------
        dt : ``float``
            Target sampling interval in sec
        """
        npts = int(self.rflength * (self.sampling / dt)) + 1
        self.data_prime = resample(self.data_prime, npts, axis=1)
        exec('self.data{} = self.data_prime'.format(self.comp.lower()))
        if not self.only_r:
            self.datat = resample(self.datat, npts, axis=1)
        self.sampling = dt
        self.rflength = npts
        self.time_axis = np.arange(npts) * dt - self.shift

    def sort(self, key='bazi'):
        """Sort RFs by keys in given ``event``, ``evla``, ``evlo``, ``evdp``,
        ``dis``, ``bazi``, ``rayp``, ``mag``, ``f0``

        :param key: key to sort, defaults to ``bazi``
        :type key: str, optional
        """
        idx = np.argsort(self.__dict__[key])
        for keyarg in self.dtype['names']:
            self.__dict__[keyarg] = self.__dict__[keyarg][idx]
        self.data_prime = self.data_prime[idx]
        exec('self.data{} = self.data_prime'.format(self.comp.lower()))
        if not self.only_r:
            self.datat = self.datat[idx]

    def moveoutcorrect(self, ref_rayp=0.06, dep_range=np.arange(0, 150), velmod='iasp91', replace=False, **kwargs):
        """Moveout correction with specified reference ray-parameter and depth

        :param ref_rayp: reference ray-parameter in s/km, defaults to 0.06
        :type ref_rayp: float, optional
        :param dep_range: Depth range used for extracting velocity in velocity model, defaults to np.arange(0, 150)
        :type dep_range: ``np.ndarray``, optional
        :param velmod: Velocity model for moveout correction. 'iasp91', 'prem' 
                      and 'ak135' is valid for internal model. Specify path to velocity model for the customized model. 
                      The format is the same as in Taup, but the depth should be monotonically increasing, defaults to 'iasp91'
        :type velmod: str, optional
        :param replace: whether replace original data, False to return new array, defaults to False
        :type replace: bool, optional

        Returns
        -------
        rf_corr: ``np.ndarray``
            Corrected RFs with component of ``RFStation.comp``

        t_corr: ``np.ndarray`` or ``None``
            Corrected RFs in transverse component. If ``only_r`` is ``True``, this variable is ``None``
        
        """
        if not self.only_r:
            t_corr, _ = moveoutcorrect_ref(self, skm2srad(ref_rayp), dep_range, chan='t', velmod=velmod, **kwargs)
        else:
            t_corr = None
        if 'datar' in self.__dict__:
            chan = 'r'
        elif 'dataz' in self.__dict__:
            chan = 'z'
        elif 'datal' in self.__dict__:
            chan = 'l'
        else:
            pass
        rf_corr, _ = moveoutcorrect_ref(self, skm2srad(ref_rayp), dep_range, chan=chan, velmod=velmod, **kwargs)
        if replace:
            self.__dict__['data{}'.format(chan)] = rf_corr
            if not self.only_r:
                self.datat = t_corr
        else:
            return rf_corr, t_corr

    def psrf2depth(self, dep_range=np.arange(0, 150), **kwargs):
        """Time-to-depth conversion with specified depth series.

        :param dep_range: Discret conversion depth, defaults to np.arange(0, 150)
        :type dep_range: :meth:`np.ndarray` , optional
        :param velmod: Velocity model for time-to-depth conversion. 'iasp91', 'prem' 
                      and 'ak135' is valid for internal model. Specify path to velocity model for the customized model. 
                      The format is the same as in Taup, but the depth should be monotonically increasing, defaults to 'iasp91'
        :type velmod: str, optional
        :param srayp: Ray-parameter lib for Ps phases, If set up to None the rayp of direct is used, defaults to None
        :type srayp: numpy.lib.npyio.NpzFile, optional
        :return: 2D array of RFs in depth
        :rtype: :meth:`np.ndarray`
        """
        self.dep_range = dep_range
        rfdepth, endindex, x_s, x_p = psrf2depth(self, dep_range, **kwargs)
        return rfdepth

    def psrf_1D_raytracing(self, dep_range=np.arange(0, 150), **kwargs):
        """1D back ray tracing to obtained Ps conversion points at discret depths

        :param dep_range: Discret conversion depth, defaults to np.arange(0, 150)
        :type dep_range: numpy.ndarray, optional
        :param velmod: Velocity model for time-to-depth conversion. 'iasp91', 'prem' 
                      and 'ak135' is valid for internal model. Specify path to velocity model for the customized model. 
                      The format is the same as in Taup, but the depth should be monotonically increasing, defaults to 'iasp91'
        :type velmod: str, optional
        :param srayp: Ray-parameter lib for Ps phases, If set up to None the rayp of direct is used, defaults to None
        :type srayp: numpy.lib.npyio.NpzFile, optional
        :return pplat_s: Latitude of conversion points
        :return pplon_s: Longitude of conversion points
        :return tps: Time difference of Ps at each depth
        :rtype: list
        """
        self.dep_range = dep_range
        pplat_s, pplon_s, _ , _, _, _, tps = psrf_1D_raytracing(self, dep_range, **kwargs)
        return pplat_s, pplon_s, tps

    def psrf_3D_raytracing(self, mod3dpath, dep_range=np.arange(0, 150), srayp=None):
        self.dep_range = dep_range
        mod3d = Mod3DPerturbation(mod3dpath, dep_range)
        pplat_s, pplon_s, _, _, tps = psrf_3D_raytracing(self, dep_range, mod3d, srayp=srayp)
        return pplat_s, pplon_s, tps

    def psrf_3D_moveoutcorrect(self, mod3dpath, **kwargs):
        warnings.warn('The fuction will be change to RFStation.psrf_3D_timecorrect in the future')
        self.psrf_3D_timecorrect(mod3dpath, **kwargs)

    def psrf_3D_timecorrect(self,  mod3dpath, dep_range=np.arange(0, 150), normalize='single', **kwargs):
        self.dep_range = dep_range
        mod3d = Mod3DPerturbation(mod3dpath, dep_range)
        pplat_s, pplon_s, pplat_p, pplon_p, raylength_s, raylength_p, tps = psrf_1D_raytracing(self, dep_range, **kwargs)
        tps = psrf_3D_migration(pplat_s, pplon_s, pplat_p, pplon_p, raylength_s, raylength_p, tps, dep_range, mod3d)
        rfdepth, _ = time2depth(self, dep_range, tps, normalize=normalize)
        return rfdepth

    def jointani(self, tb, te, tlen=3., stack_baz_val=10, rayp=0.06,
                 velmodel='iasp91', weight=[0.4, 0.4, 0.2]):
        """Eastimate crustal anisotropy with a joint method. See Liu and Niu (2012, doi: 10.1111/j.1365-246X.2011.05249.x) in detail.

        :param tb: Time before Pms for search Ps peak
        :type tb: float
        :param te: Time after Pms for search Ps peak
        :type te: float
        :param tlen: Half time length for cut out Ps phase, defaults to 3.0
        :type tlen: float, optional
        :param stack_baz_val: The interval for stacking binned by back-azimuth, defaults to 10
        :type stack_baz_val: float, optional
        :param rayp: Reference ray-parameter for moveout correction, defaults to 0.06
        :type rayp: float, optional
        :param velmodel: velocity model for moveout correction. 'iasp91', 'prem' 
                      and 'ak135' is valid for internal model. Specify path to velocity model for the customized model. 
                      The format is the same as in Taup, but the depth should be monotonically increasing, defaults to 'iasp91'
        :type velmodel: str, optional
        :param weight: Weight for three different method, defaults to [0.4, 0.4, 0.2]
        :type weight: list, optional
        :return: Dominant fast velocity direction and time delay
        :rtype: ``List``, ``List``
        """
        self.ani = RFAni(self, tb, te, tlen=tlen, rayp=rayp, model=velmodel)
        self.ani.baz_stack(val=stack_baz_val)
        best_f, best_t = self.ani.joint_ani(weight=weight)
        return best_f, best_t

    def slantstack(self, ref_dis=None, rayp_range=None, tau_range=None):
        self.slant = SlantStack(self.data_prime, self.time_axis, self.dis)
        self.slant.stack(ref_dis, rayp_range, tau_range)
        return self.slant.stack_amp

    def harmonic(self, tb=-5, te=10):
        """Harmonic decomposition for extracting anisotropic and isotropic features from the radial and transverse RFs

        :param tb: Start time relative to P, defaults to -5
        :type tb: ``float``, optional
        :param te: End time relative to P, defaults to 10
        :type te: ``float``, optional

        Returns
        -------
        harmonic_trans: numpy.ndarray, float
                Harmonic components with shape of ``(5, nsamp)``, ``nsamp = (te-tb)/RFStation.sampling``

        unmodel_trans: numpy.ndarray, float
                Unmodel components with shape same as harmonic_trans.
        """
        if self.only_r:
            raise ValueError('Transverse RFs are nessary for harmonic decomposition')
        self.harmo = Harmonics(self, tb, te)
        self.harmo.harmo_trans()
        return self.harmo.harmonic_trans, self.harmo.unmodel_trans

    def plotrt(self, outformat=None, **kwargs):
        if self.only_r:
            raise ValueError('Transverse RFs are nessary or use RFStation.plotr instead.')
        return _plotrt(self, outformat=outformat, **kwargs)

    def plotr(self, outformat=None, **kwargs):
        return _plotr(self, outformat=outformat, **kwargs)


class SACStation(RFStation):
    def __init__(self, data_path, only_r=False):
        """Class for derivative process of RFs.

        :param data_path: Path to RF data with SAC format. A finallist.dat must be in this path.
        :type data_path: str
        :param only_r: Wether only read R component, defaults to False
        :type only_r: bool, optional
        """
        super().__init__(data_path, only_r=only_r)


def _imag2nan(arr):
    StopIndex = np.where(np.imag(arr) == 1)[0]
    if StopIndex.size != 0:
        arr[StopIndex[0]:] = np.nan
    return arr


def moveoutcorrect_ref(stadatar, raypref, YAxisRange, 
                       chan='r', velmod='iasp91', sphere=True, phase=1):
    """Moveout correction refer to a specified ray-parameter
    
    :param stadatar: data class of RFStation
    :param raypref: referred ray parameter in rad
    :param YAxisRange: Depth range in nd.array type
    :param velmod: Path to velocity model
    :param chan: channel name for correction, 'r', 't'...

    :return: Newdatar, EndIndex
    """
    sampling = stadatar.sampling
    shift = stadatar.shift
    if chan == 'r':
        data = stadatar.datar
    elif chan == 't':
        data = stadatar.datat
    elif chan == 'z':
        data = stadatar.dataz
    elif chan == 'l':
        data = stadatar.datal
    else:
        raise ValueError('Field \'datar\' or \'datal\' must be in the SACStation')
    dep_mod = DepModel(YAxisRange, velmod, stadatar.stel)
    # x_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    # x_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    tps = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    for i in range(stadatar.ev_num):
        tps[i], _, _ = xps_tps_map(dep_mod, stadatar.rayp[i], stadatar.rayp[i], sphere=sphere, phase=phase)
    Tpds_ref, _, _ = xps_tps_map(dep_mod, raypref, raypref, sphere=sphere, phase=phase)
    Newdatar = np.zeros([stadatar.ev_num, stadatar.rflength])
    EndIndex = np.zeros(stadatar.ev_num)
#
    for i in range(stadatar.ev_num):
        Newaxis = np.array([])
        TempTpds = tps[i, :]
        StopIndex = np.where(np.imag(TempTpds) == 1)[0]
        if StopIndex.size == 0:
            StopIndex = dep_mod.depths.shape[0]
        EndIndex[i] = StopIndex - 1
        Newaxis = np.append(Newaxis, np.append(np.arange(-shift, 0, sampling), 0))
        for j in np.arange(int(shift / sampling + 1), stadatar.rflength):
            Refaxis = j * sampling - shift
            index = np.where(Refaxis <= tps[i, 0:StopIndex])[0]
            if index.size == 0:
                break
            Ratio = (Tpds_ref[index[0]] - Tpds_ref[index[0] - 1]) / (tps[i, index[0]] - tps[i, index[0] - 1])
            Newaxis = np.append(Newaxis, Tpds_ref[index[0] - 1] + (Refaxis - tps[i, index[0] - 1]) * Ratio)
        endidx = Newaxis.shape[0]
        x_new = np.arange(0, stadatar.rflength) * sampling - shift
        Tempdata = interp1d(Newaxis, data[i, 0:endidx], bounds_error=False)(x_new)
        endIndice = np.where(np.isnan(Tempdata))[0]
        if endIndice.size == 0:
            New_data = Tempdata
        else:
            New_data = np.append(Tempdata[1:endIndice[0]], data[i, endidx+1:])
        if New_data.shape[0] < stadatar.rflength:
            Newdatar[i] = np.append(New_data, np.zeros(stadatar.rflength - New_data.shape[0]))
        else:
            Newdatar[i] = New_data[0: stadatar.rflength]
    return Newdatar, EndIndex


def psrf2depth(stadatar, YAxisRange, velmod='iasp91', srayp=None, normalize='single', sphere=True, phase=1):
    """ Time-to-depth conversion with S-wave backprojection.

    :param stadatar: Data class of RFStation
    :type stadatar: :meth:`RFStation`
    :param YAxisRange: Depth range for conversion
    :type YAxisRange: ``numpy.ndarray``
    :param velmod: Velocity for conversion, whcih can be a path to velocity file, defaults to 'iasp91'
    :type velmod: str, optional
    :param srayp: ray-parameter library of conversion phases. See :meth:`seispy.psrayp` in detail, defaults to None
    :type srayp: str or :meth:`seispy.psrayp.PsRayp`, optional
    :param normalize: method of normalization, defaults to 'single'. Please refer to :meth:`RFStation.normalize`
    :type normalize: str, optional
    :param sphere: Wether do earth-flattening transformation, defaults to True
    :type sphere: bool, optional

    Returns
    ------------

    ps_rfdepth: 2-D numpy.ndarray, float
                RFs in depth with shape of ``(stadatar.ev_num, YAxisRange.size)``, ``stadatar.ev_num`` is the number of RFs in current station.
                ``YAxisRange.size`` is the size of depth axis.
    endindex: numpy.ndarray, int
                End index of each RF in depth
    x_s: 2-D numpy.ndarray, float
                Horizontal distance between station and S-wave conversion points with shape of  ``(stadatar.ev_num, YAxisRange.size)``
    x_p: 2-D numpy.ndarray, float
                Horizontal distance between station and P-wave conversion points with shape of  ``(stadatar.ev_num, YAxisRange.size)``
    """
    if exists(velmod):
        try:
            dep_mod = DepModel(YAxisRange, velmod, elevation=stadatar.stel)
        except:
            dep_mod = DepModel(YAxisRange, 'iasp91', elevation=stadatar.stel)
            try:
                velmod_3d = np.load(velmod)
                dep_mod.vp, dep_mod.vs = interp_depth_model(velmod_3d,
                                         stadatar.stla, stadatar.stlo, dep_mod.depths_elev)
            except Exception as e:
                raise FileNotFoundError('Cannot load 1D or 3D velocity model of \'{}\''.format(velmod))
    else:
        try:
            dep_mod = DepModel(YAxisRange, velmod, elevation=stadatar.stel)
        except:
            raise ValueError('Cannot recognize the velocity model of \'{}\''.format(velmod))

    x_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    x_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    tps = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    if srayp is None:
        for i in range(stadatar.ev_num):
            tps[i], x_s[i], x_p[i] = xps_tps_map(dep_mod, stadatar.rayp[i], 
                stadatar.rayp[i], sphere=sphere, phase=phase)
    elif isinstance(srayp, str) or isinstance(srayp, np.lib.npyio.NpzFile):
        if isinstance(srayp, str):
            if not exists(srayp):
                raise FileNotFoundError('Ps rayp lib file was not found')
            else:
                rayp_lib = np.load(srayp)
        else:
            rayp_lib = srayp
        for i in range(stadatar.ev_num):
            rayp = get_psrayp(rayp_lib, stadatar.dis[i], stadatar.evdp[i], dep_mod.depths_elev)
            rayp = skm2srad(sdeg2skm(rayp))
            tps[i], x_s[i], x_p[i] = xps_tps_map(dep_mod,
                rayp, stadatar.rayp[i], sphere=sphere, phase=phase)
    else:
        raise TypeError('srayp should be path to Ps rayp lib')
    ps_rfdepth, endindex = time2depth(stadatar, dep_mod.depths, tps, normalize=normalize)
    return ps_rfdepth, endindex, x_s, x_p


def xps_tps_map(dep_mod, srayp, prayp, is_raylen=False, sphere=True, phase=1):
    """Calculate horizontal distance and time difference at depths

    :param dep_mod: 1D velocity model class 
    :type dep_mod: :meth:`seispy.util.DepModel`
    :param srayp: conversion phase ray-parameters
    :type srayp: float
    :param prayp: S-wave ray-parameters
    :type prayp: float
    :param is_raylen: Wether calculate ray length at depths, defaults to False
    :type is_raylen: bool, optional
    :param sphere: Wether do earth-flattening transformation, defaults to True, defaults to True
    :type sphere: bool, optional
    :param phase: Phases to calculate 1 for ``Ps``, 2 for ``PpPs``, 3 for ``PsPs+PpSs``, defaults to 1
    :type phase: int, optional

    Returns
    -----------
    If ``is_raylen = False``

    tps: 2-D numpy.ndarray, float
        RFs in depth with shape of ``(stadatar.ev_num, YAxisRange.size)``, ``stadatar.ev_num`` is the number of RFs in current station.
        ``YAxisRange.size`` is the size of depth axis.
    
    x_s: 2-D numpy.ndarray, float
        Horizontal distance between station and S-wave conversion points with shape of  ``(stadatar.ev_num, YAxisRange.size)``
    x_p: 2-D numpy.ndarray, float
        Horizontal distance between station and P-wave conversion points with shape of  ``(stadatar.ev_num, YAxisRange.size)``

    otherwise, two more variables will be returned

    raylength_s: 2-D numpy.ndarray, float
    
    raylength_p: 2-D numpy.ndarray, float   
    """
    x_s = dep_mod.radius_s(prayp, phase='S', sphere=sphere)
    x_p = dep_mod.radius_s(prayp, phase='P', sphere=sphere)
    if is_raylen:
        raylength_s = dep_mod.raylength(srayp, phase='S', sphere=sphere)
        raylength_p = dep_mod.raylength(prayp, phase='P', sphere=sphere)
    if phase == 1:
        tps = dep_mod.tpds(srayp, prayp, sphere=sphere)
    elif phase == 2:
        tps = dep_mod.tpppds(srayp, prayp, sphere=sphere)
    elif phase == 3:
        tps = dep_mod.tpspds(srayp, sphere=sphere)
    else:
        raise ValueError('Phase must be in 1 for Ps, 2 for PpPs, 3 for PsPs+PpSs')
    if dep_mod.elevation != 0:
        x_s = interp1d(dep_mod.depths_elev, x_s, bounds_error=False, fill_value=(np.nan, x_s[-1]))(dep_mod.depths)
        x_p = interp1d(dep_mod.depths_elev, x_p, bounds_error=False, fill_value=(np.nan, x_p[-1]))(dep_mod.depths)
        tps = interp1d(dep_mod.depths_elev, tps, bounds_error=False, fill_value=(np.nan, tps[-1]))(dep_mod.depths)
        if is_raylen:
            raylength_s = interp1d(dep_mod.depths_elev, raylength_s, bounds_error=False, fill_value=(np.nan, raylength_s[-1]))(dep_mod.depths)
            raylength_p = interp1d(dep_mod.depths_elev, raylength_p, bounds_error=False, fill_value=(np.nan, raylength_p[-1]))(dep_mod.depths)         
    if is_raylen:
        return tps, x_s, x_p, raylength_s, raylength_p
    else:
        return tps, x_s, x_p


def psrf_1D_raytracing(stadatar, YAxisRange, velmod='iasp91', srayp=None, sphere=True, phase=1):
    dep_mod = DepModel(YAxisRange, velmod, stadatar.stel)

    # x_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    raylength_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    pplat_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    pplon_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    # x_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    raylength_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    pplat_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    pplon_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    tps = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    if srayp is None:
        for i in range(stadatar.ev_num):
            tps[i], x_s, x_p, raylength_s[i], raylength_p[i] = xps_tps_map(
                dep_mod, stadatar.rayp[i], stadatar.rayp[i], is_raylen=True, sphere=sphere, phase=phase)
            pplat_s[i], pplon_s[i] = latlon_from(stadatar.stla, stadatar.stlo, stadatar.bazi[i], rad2deg(x_s))
            pplat_p[i], pplon_p[i] = latlon_from(stadatar.stla, stadatar.stlo, stadatar.bazi[i], rad2deg(x_p))
    elif isinstance(srayp, str) or isinstance(srayp, np.lib.npyio.NpzFile):
        if isinstance(srayp, str):
            if not exists(srayp):
                raise FileNotFoundError('Ps rayp lib file not found')
            else:
                rayp_lib = np.load(srayp)
        else:
            rayp_lib = srayp
        for i in range(stadatar.ev_num):
            rayp = get_psrayp(rayp_lib, stadatar.dis[i], stadatar.evdp[i], dep_mod.depths_elev)
            rayp = skm2srad(sdeg2skm(rayp))
            tps[i], x_s, x_p, raylength_s[i], raylength_p[i] = xps_tps_map(dep_mod, rayp, 
                                        stadatar.rayp[i], is_raylen=True, sphere=sphere, phase=phase)
            x_s = _imag2nan(x_s)
            x_p = _imag2nan(x_p)
            pplat_s[i], pplon_s[i] = latlon_from(stadatar.stla, stadatar.stlo, stadatar.bazi[i], rad2deg(x_s))
            pplat_p[i], pplon_p[i] = latlon_from(stadatar.stla, stadatar.stlo, stadatar.bazi[i], rad2deg(x_p))
    else:
        raise TypeError('srayp should be path to Ps rayp lib')
    return pplat_s, pplon_s, pplat_p, pplon_p, raylength_s, raylength_p, tps


def psrf_3D_raytracing(stadatar, YAxisRange, mod3d, srayp=None, elevation=0, sphere=True):
    """
    Back ray trace the S wavs with a assumed ray parameter of P.

    :param stadatar: The data class including PRFs and more parameters
    :type stadatar: object RFStation
    :param YAxisRange: The depth array with the same intervals
    :type YAxisRange: numpy.ndarray
    :param mod3d:  The 3D velocity model with fields of ``dep``, ``lat``,
                    ``lon``, ``vp`` and ``vs``.
    :type mod3d: 'Mod3DPerturbation' object
    :param elevation: Elevation of this station relative to sea level
    :type elevation: float
    :return: pplat_s, pplon_s, pplat_p, pplon_p, tps
    :type: numpy.ndarray * 5
    """
    if sphere:
        R = 6371.0 - YAxisRange + elevation
    else:
        R = 6371.0 + elevation
    dep_range = YAxisRange.copy()
    YAxisRange -= elevation
    ddepth = np.mean(np.diff(YAxisRange))
    pplat_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    pplon_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    pplat_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    pplon_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    x_s = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    x_p = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    tps = np.zeros([stadatar.ev_num, YAxisRange.shape[0]])
    rayps = srad2skm(stadatar.rayp)

    if isinstance(srayp, str) or isinstance(srayp, np.lib.npyio.NpzFile):
        if isinstance(srayp, str):
            if not exists(srayp):
                raise FileNotFoundError('Ps rayp lib file not found')
            else:
                rayp_lib = np.load(srayp)
        else:
            rayp_lib = srayp
    elif srayp is None:
        pass
    else:
        raise TypeError('srayp should be path to Ps rayp lib')

    for i in range(stadatar.ev_num):
        if srayp is None:
            srayps = stadatar.rayp[i]
        else:
            srayps = get_psrayp(rayp_lib, stadatar.dis[i],
                                stadatar.evdp[i], YAxisRange)
            srayps = skm2srad(sdeg2skm(srayps))
        pplat_s[i][0] = pplat_p[i][0] = stadatar.stla
        pplon_s[i][0] = pplon_p[i][0] = stadatar.stlo
        x_s[i][0] = 0
        x_p[i][0] = 0
        vs = np.zeros_like(YAxisRange)
        vp = np.zeros_like(YAxisRange)
        for j, dep in enumerate(YAxisRange[:-1]):
            vs[j] = interpn((mod3d.model['dep'], mod3d.model['lat'], mod3d.model['lon']),
                            mod3d.model['vs'], (dep, pplat_s[i, j], pplon_s[i, j]),
                            bounds_error=False, fill_value=None)
            vp[j] = interpn((mod3d.model['dep'], mod3d.model['lat'], mod3d.model['lon']),
                            mod3d.model['vp'], (dep, pplat_p[i, j], pplon_p[i, j]),
                            bounds_error=False, fill_value=None)
            x_s[i, j+1] = ddepth*tand(asind(vs[j]*rayps[i])) + x_s[i, j]
            x_p[i, j+1] = ddepth*tand(asind(vp[j]*rayps[i])) + x_p[i, j]
            pplat_s[i, j+1], pplon_s[i, j+1] = latlon_from(stadatar.stla,
                                                           stadatar.stlo,
                                                           stadatar.bazi[i],
                                                           km2deg(x_s[i, j+1]))
            pplat_p[i, j+1], pplon_p[i, j+1] = latlon_from(stadatar.stla,
                                                           stadatar.stlo,
                                                           stadatar.bazi[i],
                                                           km2deg(x_p[i, j+1]))
        tps_corr = np.cumsum((np.sqrt((R / vs) ** 2 - srayps ** 2) -
                            np.sqrt((R / vp) ** 2 - stadatar.rayp[i] ** 2))
                            * (ddepth / R))
        if elevation != 0:
            tps[i] = interp1d(YAxisRange, tps_corr)(dep_range)
    return pplat_s, pplon_s, pplat_p, pplon_p, tps


def interp_depth_model(model, lat, lon, new_dep):
    """ Interpolate Vp and Vs from 3D velocity with a specified depth range.

    Parameters
    ----------
    mod3d : :meth:`np.lib.npyio.NpzFile`
        3D velocity loaded from a ``.npz`` file
    lat : float
        Latitude of position in 3D velocity model
    lon : float
        Longitude of position in 3D velocity model
    new_dep : :meth:`np.ndarray`
        1D array of depths in km

    Returns
    -------
    Vp : :meth:`np.ndarray`
        Vp in ``new_dep``
    Vs : :meth:`np.ndarray`
        Vs in ``new_dep``
    """
    #  model = np.load(modpath)
    points = [[depth, lat, lon] for depth in new_dep]
    vp = interpn((model['dep'], model['lat'], model['lon']), model['vp'], points, bounds_error=False, fill_value=None)
    vs = interpn((model['dep'], model['lat'], model['lon']), model['vs'], points, bounds_error=False, fill_value=None)
    return vp, vs


def psrf_3D_migration(pplat_s, pplon_s, pplat_p, pplon_p, raylength_s, raylength_p, Tpds, dep_range, mod3d):
    """ 3D time difference correction with specified ray path and 3D velocity model. 
        The input parameters can be generated with :meth:`psrf_1D_raytracing`.

    Parameters
    ----------
    pplat_s : :meth:`np.ndarray`
        2D array of latitude of S-wave in ``dep_range``, (:meth:`RFStation.ev_num`, ``dep_range.size``)
    pplon_s : :meth:`np.ndarray`
        2D array of longitude of S-wave in ``dep_range``, (:meth:`RFStation.ev_num`, ``dep_range.size``)
    pplat_p : :meth:`np.ndarray`
        2D array of latitude of P-wave in ``dep_range``, (:meth:`RFStation.ev_num`, ``dep_range.size``)
    pplon_p : :meth:`np.ndarray`
        2D array of longitude of P-wave in ``dep_range``, (:meth:`RFStation.ev_num`, ``dep_range.size``)
    raylength_s : :meth:`np.ndarray`
        2D array of ray path length of S-wave in ``dep_range``, (:meth:`RFStation.ev_num`, ``dep_range.size``)
    raylength_p : :meth:`np.ndarray`
        2D array of ray path length of P-wave in ``dep_range``, (:meth:`RFStation.ev_num`, ``dep_range.size``)
    Tpds : :meth:`np.ndarray`
        1D array of time difference in ``dep_range`` (``dep_range.size``)
    dep_range : :meth:`np.ndarray`
        1D array of depths in km, (``dep_range.size``)
    mod3d : :meth:`np.lib.npyio.NpzFile`
        3D velocity loaded from a ``.npz`` file

    Returns
    -------
    :meth:`np.ndarray`
        Corrected time difference in dep_range
    """
    ev_num, _ = raylength_p.shape
    timecorrections = np.zeros_like(raylength_p)
    for i in range(ev_num):
        points = np.array([dep_range, pplat_p[i], pplon_p[i]]).T
        dvp = mod3d.interpdvp(points)
        points = np.array([dep_range, pplat_s[i], pplon_s[i]]).T
        dvs = mod3d.interpdvs(points)
        dlp = raylength_p[i]
        dls = raylength_s[i]
        tmpds = (dls / (mod3d.cvs * (1 + dvs)) - dls / mod3d.cvs) - (dlp / (mod3d.cvp * (1 + dvp)) - dlp / mod3d.cvp)
        tmpds[np.isnan(tmpds)] = 0
        timecorrections[i] = np.cumsum(tmpds)
    return Tpds + timecorrections


def time2depth(stadatar, dep_range, Tpds, normalize='single'):
    """ Interpolate RF amplitude with specified time difference and depth range

    Parameters
    ----------
    stadatar : :meth:`RFStation` 
        Data class of :meth:`RFStation` 
    dep_range : :meth:`np.ndarray`
        1D array of depths in km, (``dep_range.size``)
    Tpds : :meth:`np.ndarray`
        1D array of time difference in ``dep_range`` (``dep_range.size``)
    normalize : str, optional
        Normlization option, ``'sinlge'`` and ``'average'`` are available , by default 'single'
        See :meth:`RFStation.normalize` in detail.

    Returns
    -------
    PS_RFdepth: :meth:`np.ndarray`
        2D array of RFs in depth, (:meth:`RFStation.ev_num`, ``dep_range.size``)
    end_index: :meth:`np.ndarray`
        1D array of the last digit of effective depth (:meth:`RFStation.ev_num`)
    """
    if normalize:
        stadatar.normalize(method=normalize)
    PS_RFdepth = np.zeros([stadatar.ev_num, dep_range.shape[0]])
    EndIndex = np.zeros(stadatar.ev_num).astype(int)
    for i in range(stadatar.ev_num):
        TempTpds = Tpds[i, :]
        StopIndex = np.where(np.imag(TempTpds) == 1)[0]
        if StopIndex.size == 0:
            EndIndex[i] = dep_range.size - 1
            # DepthAxis = interp1d(TempTpds, dep_range, bounds_error=False)(stadatar.time_axis)
        else:
            EndIndex[i] = StopIndex[0] - 1
            # DepthAxis = interp1d(TempTpds[0:StopIndex], dep_range[0: StopIndex], bounds_error=False)(stadatar.time_axis)

        PS_RFTempAmps = stadatar.__dict__['data{}'.format(stadatar.comp.lower())][i]
        ValueIndices = np.where(np.logical_not(np.isnan(dep_range[0:EndIndex[i]+1])))[0]
        if ValueIndices.size == 0:
            continue
        elif np.max(ValueIndices) > PS_RFTempAmps.shape[0]:
            continue
        else:
            PS_RFAmps = interp1d(stadatar.time_axis, PS_RFTempAmps, bounds_error=False)(TempTpds[0:EndIndex[i]+1])
            PS_RFdepth[i] = PS_RFAmps
    return PS_RFdepth, EndIndex


if __name__ == '__main__':
    rfsta = SACStation('/Users/xumijian/Codes/seispy-example/ex-ccp/RFresult/ZX.212')
    rfsta.jointani(2, 7, weight=[0.9, 0.1, 0.0])
    rfsta.ani.plot_polar()




#!/usr/bin/env python
from setuptools import find_packages, setup
packages = find_packages()

with open("README.md", "r") as fh:
    long_description = fh.read()


VERSION = "1.3.3"
setup(name='python-seispy',
      version=VERSION,
      author='Mijian Xu',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author_email='gomijianxu@gmail.com',
      license='GPLv3',
      packages=find_packages(),
      package_dir={'seispy': 'seispy'},
      package_data={'': ['data/*']},
      install_requires=[
                'numpy>=1.19.0',
                'scipy>=1.9.1',
                'matplotlib>=3.5.0',
                'pandas>=1.0.0',
                'obspy>=1.2.1',
                'pyside6>=6.2.0',
                'scikits.bootstrap>=1.0.0'],
      entry_points={'console_scripts': ['gen_rayp_lib=seispy.psrayp:gen_rayp_lib',
                                        'prf=seispy.scripts:prf',
                                        'srf=seispy.scripts:srf',
                                        'setpar=seispy.rf:setpar',
                                        'rf2depth=seispy.rf2depth_makedata:rf2depth',
                                        'plotrt=seispy.scripts:plot_rt',
                                        'plotr=seispy.scripts:plot_r',
                                        'download_catalog=seispy.catalog:main',
                                        'ccp_profile=seispy.scripts:ccp_profile',
                                        'hk=seispy.hk:hk',
                                        'pickrf=seispy.pickrf.pickui:main',
                                        'pickdepth=seispy.pickdepth.pickdepthui:main',
                                        'rfani=seispy.scripts:rfani',
                                        'ccp3d=seispy.scripts:ccp3d',
                                        'rfharmo=seispy.scripts:rfharmo',
                                        'get_pierce_points=seispy.scripts:get_pierce_points',
                                        'veltxt2mod=seispy.modcreator:veltxt2mod']},
      zip_safe=False,
      classifiers=['Programming Language :: Python',
                   'Programming Language :: Python :: 3.9',
                   'Programming Language :: Python :: 3.10',
                   'Programming Language :: Python :: 3.11']
      )



import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
from obspy.taup import TauPyModel
from seispy.geo import srad2skm, skm2sdeg

class SlantStack():
    def __init__(self, seis, timeaxis, dis) -> None:
        """Slant stack in tau domain. Refer to Tauzin et al., 2008 JGR in detail.

        :param seis: 2D array for RFs with shape of (ev_num, npts)
        :type seis: numpy.ndarray
        :param timeaxis: 1D array for time axis
        :type timeaxis: numpy.ndarray
        :param dis: 1D array for all event distance in km
        :type dis: numpy.ndarray
        """
        self.datar = seis
        self.time_axis = timeaxis
        self.dis = dis
        self.ref_dis = 65
        self.rayp_range = np.arange(-0.35, 0.35, 0.01)
        self.tau_range = np.arange(0, 100, 0.1)
        self.syn_tau = np.array([])
        self.syn_drayp = np.array([])

    def stack(self, ref_dis=None, rayp_range=None, tau_range=None):
        if ref_dis is not None and isinstance(ref_dis, (int, float)):
            self.ref_dis = ref_dis
        elif ref_dis is None:
            pass
        else:
            raise TypeError('{} should be in int or float type.'.format(ref_dis))
        if rayp_range is not None and isinstance(rayp_range, np.ndarray):
            self.rayp_range = rayp_range
        elif rayp_range is None:
            pass
        else:
            raise TypeError('{} should be in numpy.ndarray type.'.format(rayp_range))
        if tau_range is not None and isinstance(tau_range, np.ndarray):
            self.tau_range = tau_range
        elif tau_range is None:
            pass
        else:
            raise TypeError('{} should be in numpy.ndarray type.'.format(tau_range))
        ev_num = self.datar.shape[0]
        taus, rayps = np.meshgrid(self.tau_range, self.rayp_range)
        self.stack_amp = np.zeros([self.rayp_range.shape[0], self.tau_range.shape[0]])
        for i in range(ev_num):
            tps = taus - rayps * (self.dis[i] - self.ref_dis)
            self.stack_amp += interp1d(self.time_axis, self.datar[i, :], fill_value='extrapolate')(tps)
        self.stack_amp /= ev_num

    def syn_tps(self, phase_list, velmodel='iasp91', focal_dep=10):
        model = TauPyModel(model=velmodel)
        phase_list.insert(0, 'P')
        arrs = model.get_travel_times(focal_dep, self.ref_dis, phase_list=phase_list)
        p_arr = arrs[0].time
        p_rayp = skm2sdeg(srad2skm(arrs[0].ray_param))
        self.syn_tau = [arr.time - p_arr for arr in arrs[1:]]
        self.syn_drayp = [p_rayp - skm2sdeg(srad2skm(arr.ray_param))for arr in arrs[1:]]

    def plot(self, cmap='jet', xlim=None, vmin=None, vmax=None, figpath=None, colorbar=True):
        """ Imaging for slant stacking

        Parameters
        ----------
        cmap : str, optional
            The colormap to use, by default 'jet'
        xlim : _type_, optional
             Set the x limits of the current axes., by default None
        vmin : _type_, optional
            Normalize the minium value data to the ``vmin``, by default None
        vmax : _type_, optional
            Normalize the maximum value data to the ``vmax``, by default None
        figpath : _type_, optional
            Output path to the image, by default None
        colorbar : bool, optional
            Wether plot the colorbar, by default True
        """
        plt.style.use("bmh")
        self.fig = plt.figure(figsize=(8,5))
        self.ax = self.fig.add_subplot()
        if vmin is None and vmax is None:
            vmax = np.max(np.abs(self.stack_amp))*0.1
            vmin = -vmax
        elif vmin is None and isinstance(vmax, (int, float)):
            vmin = np.min(self.stack_amp)
        elif vmax is None and isinstance(vmin, (int, float)):
            vmax = np.max(self.stack_amp)
        elif isinstance(vmax, (int, float)) and isinstance(vmin, (int, float)):
            pass
        else:
            raise TypeError('vmin and vmax must be in int or in float type')
        im = self.ax.pcolor(self.tau_range, self.rayp_range, self.stack_amp,
                            vmax=vmax, vmin=vmin, cmap=cmap)
        if colorbar:
            cax = self.fig.colorbar(im, ax=self.ax)
            cax.set_label('Stack Amplitude')
        self.ax.grid(visible=True)
        self.ax.scatter(self.syn_tau, self.syn_drayp, color='k', marker='x')
        if xlim is not None and isinstance(xlim, (list, np.ndarray)):
            self.ax.set_xlim(xlim)
        self.ax.set_xlabel('Time (s)')
        self.ax.set_ylabel('Slowness (s/$^\circ$)')
        if figpath is not None and isinstance(figpath, str):
            self.fig.savefig(figpath, format='png', dpi=400, bbox_inches='tight')
        else:
            plt.show()



import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.lines import Line2D
import numpy as np
from os.path import join


def init_figure():
    h = plt.figure(figsize=(8, 10))
    h.tight_layout()
    gs = GridSpec(15, 3)
    gs.update(wspace=0.25)
    axr = plt.subplot(gs[1:, 0:-1])
    axr.grid(color='gray', linestyle='--', linewidth=0.4, axis='x')
    axb = plt.subplot(gs[1:, -1])
    axb.grid(color='gray', linestyle='--', linewidth=0.4, axis='x')
    axs = plt.subplot(gs[0, 0:-1])
    axs.grid(color='gray', linestyle='--', linewidth=0.4, axis='x')
    return h, axr, axb, axs


def plot_waves(axr, axb, axs, stadata, enf=12):
    bound = np.zeros(stadata.rflength)
    for i in range(stadata.ev_num):
        datar = stadata.data_prime[i] * enf + (i + 1)
        axr.fill_between(stadata.time_axis, datar, bound + i+1, where=datar > i+1, facecolor='red',
                         alpha=0.7)
        axr.fill_between(stadata.time_axis, datar, bound + i+1, where=datar < i+1, facecolor='blue',
                         alpha=0.7)
    # rfcorr, _ = stadata.moveoutcorrect( 0.1, np.arange(300), sphere=False)
    # rfsum = np.mean(rfcorr, axis=0)
    rfsum = np.mean(stadata.data_prime, axis=0)
    axs.fill_between(stadata.time_axis, rfsum, 0, where=rfsum > 0, facecolor='red', alpha=0.7)
    axs.fill_between(stadata.time_axis, rfsum, 0, where=rfsum < 0, facecolor='blue', alpha=0.7)
    # axs.plot(stadata.time_axis, rfsum, color='gray', lw=0.5)
    axb.scatter(stadata.bazi, np.arange(stadata.ev_num) + 1, s=7)
    # axp = axb.twiny()
    # axp.scatter(stadata.rayp, np.arange(stadata.ev_num) + 1, s=7)
    # return axp

def set_fig(axr, axb, axs, stadata, xmin=-2, xmax=80):
    y_range = np.arange(stadata.ev_num) + 1
    x_range = np.arange(0, xmax+2, 5)
    space = 2

    # set axr
    axr.set_xlim(xmin, xmax)
    axr.set_xticks(x_range)
    axr.set_xticklabels(x_range, fontsize=8)
    axr.set_ylim(0, stadata.ev_num + space)
    axr.set_yticks(y_range)
    axr.set_yticklabels(stadata.event, fontsize=5)
    axr.set_xlabel('Time after P (s)', fontsize=13)
    axr.set_ylabel('Event', fontsize=13)
    axr.add_line(Line2D([0, 0], axr.get_ylim(), color='black'))
    # axr.set_title('R components ({})'.format(stadata.staname), fontsize=16)

    # set axb
    axb.set_xlim(0, 360)
    axb.set_xticks(np.linspace(0, 360, 7))
    axb.set_xticklabels(np.linspace(0, 360, 7, dtype='i'), fontsize=8)
    axb.set_ylim(0, stadata.ev_num + space)
    axb.set_yticks(y_range)
    axb.set_yticklabels(y_range, fontsize=5)
    axb.set_xlabel(r'Back-azimuth ($^\circ$)', fontsize=13)
    # axp.set_xlabel('Ray-parameter (s/km)', fontsize=13)

    axs.set_title('{} components ({})'.format(stadata.comp.upper(), stadata.staname), fontsize=16)
    axs.set_xlim(xmin, xmax)
    axs.set_xticks(x_range)
    axs.set_xticklabels([])
    axs.set_yticks([np.sum(axs.get_ylim())/3])
    axs.tick_params(axis='y', left=False)
    axs.set_yticklabels(['Sum'], fontsize=8)
    axs.add_line(Line2D([0, 0], axs.get_ylim(), color='black'))


def plotr(rfsta, out_path='./', xlim=[-2, 80], key='bazi', enf=6, outformat='g', show=False):
    h, axr, axb, axs = init_figure()
    rfsta.sort(key)
    plot_waves(axr, axb, axs, rfsta, enf=enf)
    set_fig(axr, axb, axs, rfsta, xlim[0], xlim[1])
    if outformat is None and not show:
        return h
    elif outformat == 'g':
        h.savefig(join(out_path, rfsta.staname+'_R_{}order_{:.1f}.png'.format(key, rfsta.f0[0])),
                       dpi=400, bbox_inches='tight')
    elif outformat == 'f':
        h.savefig(join(out_path, rfsta.staname+'_R_{}order_{:.1f}.pdf'.format(key, rfsta.f0[0])),
                       format='pdf', bbox_inches='tight')
    if show:
        plt.show()
        return h


if __name__ == '__main__':
    pass


from seispy.rf import RF, datestr2regex, SACFileNotFoundError
from seispy.eq import EQ
import numpy as np
from obspy import UTCDateTime
from datetime import timedelta
import pandas as pd
import glob
from os.path import join
import re
import sys
import obspy


def match_eq(eq_lst, pathname, logger, ref_comp='Z', suffix='SAC', offset=0,
             tolerance=1, dateformat='%Y.%j.%H.%M.%S'):
    pattern = datestr2regex(dateformat)
    ref_eqs = glob.glob(join(pathname, '*{0}*{1}'.format(ref_comp, suffix)))
    if len(ref_eqs) == 0:
        raise SACFileNotFoundError(join(pathname, '*{0}*{1}'.format(ref_comp, suffix)))
    sac_files = []
    for ref_sac in ref_eqs:
        try:
            datestr = re.findall(pattern, ref_sac)[0]
        except IndexError:
            raise IndexError('Error data format of {} in {}'.format(dateformat, ref_sac))
        if isinstance(offset, (int, float)):
            sac_files.append([datestr, UTCDateTime.strptime(datestr, dateformat), -offset])
        elif offset is None:
            try:
                tr = obspy.read(ref_sac)[0]
            except TypeError:
                continue
            sac_files.append([datestr, tr.stats.starttime-tr.stats.sac.b, tr.stats.sac.o])
        else:
            raise TypeError('offset should be int or float type')
    new_col = ['data', 'datestr']
    eq_match = pd.DataFrame(columns=new_col)
    for datestr, b_time, offs in sac_files:
        date_range_begin = b_time + timedelta(seconds=offs - tolerance)
        date_range_end = b_time + timedelta(seconds=offs + tolerance)
        results = eq_lst[(eq_lst.date > date_range_begin) & (eq_lst.date < date_range_end)]
        if len(results) != 1:
            continue
        try:
            this_eq = EQ(pathname, datestr, suffix)
        except Exception as e:
            logger.RFlog.error(''.format(e))
            continue
        this_eq.get_time_offset(results.iloc[0]['date'])
        this_df = pd.DataFrame([[this_eq, datestr]], columns=new_col, index=results.index.values)
        eq_match = pd.concat([eq_match, this_df])
    ind = eq_match.index.drop_duplicates(keep=False)
    eq_match = eq_match.loc[ind]
    return pd.concat([eq_lst, eq_match], axis=1, join='inner')


class ReRF(RF):
    def __init__(self, finallist, cfg_file=None, log=None):
        super().__init__(cfg_file, log)
        self.eq_lst = self.read_finallist(finallist)
    
    def read_finallist(self, finallist):
        """Read final list from RF path.

        Parameters
        ----------
        finallist : str
            Path to final list.

        Returns
        -------
        pandas.DataFrame
            Data frame of earthquake information.
        """
        self.logger.RFlog.info('Read event info from {}'.format(finallist))
        dtype = {'names': ('evt', 'phase', 'evlat', 'evlon', 'evdep', 'dis', 'bazi', 'rayp', 'mag', 'f0'),
                 'formats': ('U20', 'U20', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4')}
        self.event, self.phase, self.evla, self.evlo, self.evdp, self.dis, self.bazi, self.rayp, self.mag, self.f0 = \
            np.loadtxt(finallist, dtype=dtype, unpack=True, ndmin=1)
        cols = ['date', 'evla', 'evlo', 'evdp', 'dis', 'bazi', 'rayp', 'mag', 'magtype']
        lst = []
        for i, evt_str in enumerate(self.event):
            evtime = UTCDateTime.strptime(evt_str, '%Y.%j.%H.%M.%S')
            lst.append([evtime, self.evla[i], self.evlo[i], self.evdp[i],
                       self.dis[i], self.bazi[i], self.rayp[i], self.mag[i], 'mw'])
        return pd.DataFrame(lst, columns=cols)
    
    def match_eq(self, **kwarg):
        try:
            self.logger.RFlog.info('Match SAC files')
            self.eqs = match_eq(self.eq_lst, self.para.datapath, self.logger,
                                ref_comp=self.para.ref_comp, suffix=self.para.suffix,
                                offset=self.para.offset, tolerance=self.para.tolerance,
                                dateformat=self.para.dateformat)
        except Exception as e:
            self.logger.RFlog.error('{0}'.format(e))
            raise e
        if self.eqs.shape[0] == 0:
            self.logger.RFlog.warning('No earthquakes matched, please check configurations.'.format(self.eqs.shape[0]))
            sys.exit(1)
        else:
            self.logger.RFlog.info('{0} earthquakes are matched'.format(self.eqs.shape[0]))

    def write_list(self):
        path = join(self.para.rfpath, '{}.{}finallist.dat'.format(self.stainfo.network, self.stainfo.station))
        self.logger.RFlog.info('Writting event info to {}'.format(path))
        with open(path, 'w') as f:
            for i, row in self.eqs.iterrows():
                f.write('{} {} {:.3f} {:.3f} {:.3f} {:.3f} {:.3f} {:.5f} {:.3f} {:.3f}\n'.format(
                    row['date'].strftime('%Y.%j.%H.%M.%S'), self.para.phase, row['evla'], row['evlo'],
                    row['evdp'], row['dis'], row['bazi'], row['rayp'], row['mag'], self.para.gauss
                ))



import glob
import obspy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from seispy.pickrf.pickfigure import RFFigure, indexpags
from seispy.rfcorrect import RFStation
from os.path import join, basename


class RPickFigure(RFFigure):
    def __init__(self, rfpath, width=8, height=10, dpi=100, xlim=[-2, 85]):
        super().__init__(rfpath, width=width, height=height, dpi=dpi, xlim=xlim)
        self.enf = 8
        
    def init_canvas(self, order='baz'):
        return super().init_canvas(order=order)

    def init_figure(self, width, height, dpi):
        self.fig = Figure(figsize=(width, height), dpi=dpi) 
        self.axr = self.fig.add_axes([0.15, 0.05, 0.55, 0.84])
        self.axb = self.fig.add_axes([0.755, 0.05, 0.22, 0.84])
        self.axg = self.axb.twiny()
    
    def read_sac(self, dt=0.1, order='baz'):
        if not isinstance(order, str):
            raise TypeError('The order must be str type')
        elif not order in ['baz', 'dis']:
            raise ValueError('The order must be \'baz\' or \'dis\'')
        else:
            pass
        if len(glob.glob(join(self.rfpath, '*P_R.sac'))) != 0:
            tmp_files = glob.glob(join(self.rfpath, '*_R.sac'))  
            self.comp = 'R'
        elif len(glob.glob(join(self.rfpath, '*P_Q.sac'))) != 0:
            tmp_files = glob.glob(join(self.rfpath, '*P_Q.sac'))
            self.comp = 'Q'
        elif len(glob.glob(join(self.rfpath, '*S_L.sac'))) != 0:
            tmp_files = glob.glob(join(self.rfpath, '*S_L.sac'))
            self.comp = 'L'
            self.enf = 4
        elif len(glob.glob(join(self.rfpath, '*S_Z.sac'))) != 0:
            tmp_files = glob.glob(join(self.rfpath, '*S_Z.sac'))
            self.comp = 'Z'
            self.enf = 4
        else:
            self.log.RFlog.info('No such file in the format of evt_phase_component.sac')
            exit()
        self.log.RFlog.info('Reading PRFs from {}'.format(self.rfpath))
        self.filenames = [basename(sac_file).split('_')[0] for sac_file in sorted(tmp_files)]
        self.phases = [basename(sac_file).split('_')[1] for sac_file in sorted(tmp_files)]
        self.rrf = obspy.read(join(self.rfpath, '*_{}.sac'.format(self.comp))).sort(['starttime']).resample(1/dt, window=None)
        self.time_axis = self.rrf[0].times() + self.rrf[0].stats.sac.b
        if self.xlim[1] > self.time_axis[-1]:
            self.xlim[1] = self.time_axis[-1]
        self.evt_num = len(self.rrf)
        self.log.RFlog.info('A total of {} PRFs loaded'.format(self.evt_num))
        self.baz = np.array([tr.stats.sac.baz for tr in self.rrf])
        self.gcarc = np.array([tr.stats.sac.gcarc for tr in self.rrf])
        self.rayp = np.array([tr.stats.sac.user0 for tr in self.rrf])
        self._sort(order)
        self.axpages, self.rfidx = indexpags(self.evt_num, self.maxidx)
        self.staname = (self.rrf[0].stats.network+'.'+self.rrf[0].stats.station).strip('.')
        self.fig.suptitle("%s (Latitude: %5.2f\N{DEGREE SIGN}, Longitude: %5.2f\N{DEGREE SIGN})" % (self.staname, self.rrf[0].stats.sac.stla, self.rrf[0].stats.sac.stlo), fontsize=20)

    def _sort(self, order):
        if order == 'baz':
            idx = np.argsort(self.baz)
        elif order == 'dis':
            idx = np.argsort(self.gcarc)
        else:
            pass
        self.baz = self.baz[idx]
        self.gcarc = self.gcarc[idx]
        self.phases = [self.phases[i] for i in idx]
        self.rrf = obspy.Stream([self.rrf[i] for i in idx])
        # self.gcarc = [self.rrf[i].stats.sac.gcarc for i in range(self.evt_num)]
        self.filenames = [self.filenames[i] for i in idx]

    def set_figure(self):
        self.axr.grid(visible=True, which='major', axis='x')
        self.axb.grid(visible=True, which='major')
        self.axr.set_ylabel("Event")
        self.axr.set_xlabel("Time after P (s)")
        self.axr.set_title("{} component".format(self.comp))
        self.axb.set_xlabel("Backazimuth (\N{DEGREE SIGN})")
        self.axg.set_xlabel('Distance (\N{DEGREE SIGN})')

    def set_page(self):
        self.set_ylabels()
        self.axr.plot([0, 0], [0, self.axpages*self.maxidx+1], color="black")
        self.axr.set_xlim(self.xlim[0], self.xlim[1])
        self.axr.set_xticks(np.arange(self.xlim[0], self.xlim[1]+1, 2))
        self.set_ax_baz_dis()
    
    def set_ylabels(self):
        self.axr.set_ylim(self.rfidx[self.ipage][0], self.rfidx[self.ipage][0]+self.maxidx+1)
        self.axr.set_yticks(np.arange(self.rfidx[self.ipage][0], self.rfidx[self.ipage][0]+self.maxidx+1))
        ylabels = np.array(self.filenames)[self.rfidx[self.ipage]]
        ticklabels = [''] * (self.maxidx+1)        
        ticklabels[1: len(ylabels)+1] = ylabels
        self.axr.set_yticklabels(ticklabels)
        self.axb.set_ylim(self.axr.get_ylim())
        self.axb.set_yticks(self.axr.get_yticks())
        self.azi_label = ['%5.2f' % self.baz[i] for i in self.rfidx[self.ipage]]
        ticklabels = [''] * (self.maxidx+1)
        ticklabels[1: len(self.azi_label)+1] = self.azi_label
        self.axb.set_yticklabels(ticklabels)
    
    def init_variables(self):
        self.goodrf = np.ones(self.evt_num)
        self.rlines = [[]]*self.evt_num
        self.rwvfillpos = [[]]*self.evt_num
        self.rwvfillnag = [[]]*self.evt_num
    
    def plotwave(self):
        bound = np.zeros(self.time_axis.shape[0])
        for i in range(self.evt_num):
            r_amp_axis = self.rrf[i].data*self.enf+i+1
            self.rlines[i], = self.axr.plot(self.time_axis, r_amp_axis, color="black", linewidth=0.2)            
            self.rwvfillpos[i] = self.axr.fill_between(self.time_axis, r_amp_axis, bound+i+1, where=r_amp_axis >i+1, facecolor='red', alpha=0.3)
            self.rwvfillnag[i] = self.axr.fill_between(self.time_axis, r_amp_axis, bound+i+1, where=r_amp_axis <i+1, facecolor='blue', alpha=0.3)

    def onclick(self, event):
        if event.inaxes != self.axr:
            return
        click_idx = int(np.round(event.ydata))
        if click_idx > self.evt_num:
            return
        if self.goodrf[click_idx-1] == 1:
            self.log.RFlog.info("Selected "+self.filenames[click_idx-1])
            self.goodrf[click_idx-1] = 0
            self.rwvfillpos[click_idx-1].set_facecolor('gray')
            self.rwvfillnag[click_idx-1].set_facecolor('gray')
        else:
            self.log.RFlog.info("Canceled "+self.filenames[click_idx-1])
            self.goodrf[click_idx-1] = 1
            self.rwvfillpos[click_idx-1].set_facecolor('red')
            self.rwvfillnag[click_idx-1].set_facecolor('blue')

    def _set_gray(self):
        for i in np.where(self.goodrf == 0)[0]:
            self.rwvfillpos[i].set_facecolor('gray')
            self.rwvfillnag[i].set_facecolor('gray')
    
    def enlarge(self):
        self.enf += 1
        self.axr.cla()
        self.plotwave()
        self._set_gray()
        self.set_page()
        self.set_figure()

    def reduce(self):
        if self.enf > 1:
            self.enf -= 1
        else:
            self.enf = 1/(1/self.enf + 1)
        self.axr.cla()
        self.plotwave()
        self._set_gray()
        self.set_page()
        self.set_figure()

    def plot(self):
        plt.ion()
        plt.rcParams['toolbar'] = 'None'
        goodidx = np.where(self.goodrf == 1)[0]
        newrfs = obspy.Stream([self.rrf[idx] for idx in goodidx])
        stadata = RFStation.read_stream(newrfs, self.rayp[goodidx],
                                        self.baz[goodidx], prime_comp=self.comp)
        stadata.event = np.array([self.filenames[i] for i in goodidx])
        self.plotfig = stadata.plotr(enf=self.enf, xlim=self.xlim)


import numpy as np
from scipy.sparse.linalg import lsqr
from seispy.geo import cosd, sind
import matplotlib.pyplot as plt
from obspy.io.sac import SACTrace
from os.path import join



class Harmonics():
    def __init__(self, rfsta, tmin=-5, tmax=10) -> None:
        """ Harmonic decomposition for extracting anisotropic and isotropic features from the radial and transverse RFs

        :param rfsta: data class of RFStation
        :type rfsta: :class:`seispy.rfcorrect.RFStation`
        :param tmin: Start time relative to P
        :type tmin: float
        :param tmax: End time relative to P
        :type tmax: float
        """
        self.rfsta = rfsta
        self.tmin = tmin
        self.tmax = tmax
        self.cut_trace()

    def cut_trace(self):
        """Trim RFs from tmin to tmax
        """
        nb = int((self.tmin + self.rfsta.shift) / self.rfsta.sampling)
        ne = int((self.tmax + self.rfsta.shift) / self.rfsta.sampling)
        self.datar = self.rfsta.datar[:, nb:ne+1]
        self.datat = self.rfsta.datat[:, nb:ne+1]
        self.nsamp = ne - nb + 1
        self.time_axis = np.linspace(self.tmin, self.tmax, self.nsamp)

    def harmo_trans(self):
        self.traces = np.vstack((self.datar, self.datat))
        harmonic = np.zeros((self.rfsta.ev_num*2, 5))
        unmodel = np.zeros((self.rfsta.ev_num*2, 5))
        self.harmonic_trans = np.zeros((5, self.nsamp))
        self.unmodel_trans = np.zeros((5, self.nsamp))
        for i, baz in enumerate(self.rfsta.bazi):
            harmonic[i] = np.array([1, cosd(baz), sind(baz), cosd(baz*2), sind(baz*2)])
            harmonic[i+self.rfsta.ev_num] = np.array([0, cosd(baz+90), sind(baz+90), cosd(baz*2+90), sind(baz*2+90)])
            unmodel[i] = np.array([1, cosd(baz), sind(baz), cosd(baz*2), sind(baz*2)])
            unmodel[i+self.rfsta.ev_num] = np.array([0, cosd(baz-90), sind(baz-90), cosd(baz*2-90), sind(baz*2-90)])
        for i in range(self.nsamp):
            b = self.traces[:, i]
            self.harmonic_trans[:, i] = lsqr(harmonic, b, damp=0)[0]
            self.unmodel_trans[:, i] = lsqr(unmodel, b, damp=0)[0]
    
    def write_constant(self, out_sac_path='./'):
        """ Write constant component to SAC file.

        :param out_sac_path: Output path, defaults to './'
        :type out_sac_path: str, optional
        """
        sac = SACTrace(data=self.harmonic_trans[0, :])
        sac.b = self.tmin
        sac.delta = self.rfsta.sampling
        sac.user0 = 0.0
        sac.user1 = self.rfsta.f0[0]
        sac.stla = self.rfsta.stla
        sac.stlo = self.rfsta.stlo
        sac.stel = self.rfsta.stel
        sac.write(join(out_sac_path, '{}_constant_R.sac'.format(self.rfsta.staname)))

    def plot(self, outpath='./', enf=2.):
        """Plot harmonic and unmodeled components

        :param outpath: Output path, defaults to './'
        :type outpath: str, optional
        :param enf: Amplification factor, defaults to 2.0
        :type enf: float, optional
        """
        plt.style.use("bmh")
        plt.rc('grid', color='white', linestyle='-', linewidth=0.7)
        plt.rcParams["axes.grid.axis"] = "x"
        fig, axes = plt.subplots(1, 2, figsize=(10, 5), sharey=True)
        mtx = [self.harmonic_trans, self.unmodel_trans]
        bound = np.zeros(self.nsamp)
        titles = ['Dipping/Anisotropic', 'Unmodeled']
        for iax, ax in enumerate(axes):
            for i in range(5):
                data = mtx[iax][5-i-1] * enf + (i + 1)
                ax.fill_between(self.time_axis, data, bound + i+1, where=data > i+1, facecolor='red',
                                alpha=0.7)
                ax.fill_between(self.time_axis, data, bound + i+1, where=data < i+1, facecolor='#1193F4',
                                alpha=0.7)
            ax.plot([0, 0], [0, 6], color='k', lw=1.2)
            ax.set_xlim([-1, self.tmax])
            ax.set_xlabel('Time after P (s)', fontsize=13)
            ax.set_ylim([0, 6])
            ax.set_title(titles[iax])
        axes[0].set_yticks([1, 2, 3, 4, 5])
        axes[0].set_yticklabels(['sin2${\\theta}$', 'cos2${\\theta}$', 'sin${\\theta}$', 'cos${\\theta}$', 'Constant'], fontsize=13)
        fig.savefig(join(outpath, '{}_harmonic_trans.png'.format(self.rfsta.staname)), dpi=300, bbox_inches='tight')
